# Generated by Django 4.2.23 on 2025-12-10 10:58

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0034_workflow_template_configure_runtime_parameters'),
    ]

    operations = [
        migrations.AlterField(
            model_name='workrequest',
            name='status',
            field=models.CharField(
                choices=[
                    ('pending', 'Pending'),
                    ('running', 'Running'),
                    ('completed', 'Completed'),
                    ('aborted', 'Aborted'),
                    ('blocked', 'Blocked'),
                ],
                default='blocked',
                editable=False,
                max_length=9,
            ),
        ),
    ]
