# Generated by Django 4.2.21 on 2025-09-18 16:26

from django.db import migrations, models

import debusine.db.models.work_requests


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0018_alter_assetusagerole_role'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='workflowtemplate',
            constraint=models.CheckConstraint(
                check=models.Q(
                    models.Q(('task_name', 'update_suites'), _negated=True),
                    (
                        'name',
                        debusine.db.models.work_requests.WellKnownWorkflowTemplate[
                            'UPDATE_SUITES'
                        ],
                    ),
                    _connector='OR',
                ),
                name='db_workflowtemplate_update_suites',
            ),
        ),
    ]
