{% load debusine %}
<div {% if formatted_content_div_id %}id="{{ formatted_content_div_id }}"{% endif %}
     class="card">
    <div class="card-header d-flex">
        <div class="me-auto">{{ file_in_artifact.path }}</div>
        {% if not file_in_artifact.complete %}
            <div class="badge bg-warning rounded-pill">incomplete</div>
        {% else %}
            <div class="btn-group btn-group-sm">
                <a class="btn btn-debusine"
                   href="{{ file_in_artifact.get_absolute_url_raw }}"><i title="View raw" class="bi {% icon "file_view_raw" %}"></i></a>
                <a class="btn btn-debusine"
                   href="{{ file_in_artifact.get_absolute_url_download }}"><i title="Download" class="bi {% icon "file_download" %}"></i></a>
            </div>
        {% endif %}
    </div>
    {% if file_in_artifact.complete %}
        <div class="card-body text-white bg-dark rounded-bottom ps-0 pt-1 pb-0">
            {% block formatted_content %}{% endblock %}
        </div>
    {% endif %}
</div>
