# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine task input fields."""

from debusine.tasks.inputs.inputs import (
    BuildArchitectureInput,
    DebusineFQDNInput,
    EnvironmentInput,
    ExtraRepositoriesInput,
    MultiInput,
    MultiInputList,
    OptionalEnvironmentInput,
    OptionalSingleInput,
    SingleInput,
    SingleInputList,
    Stage,
    SuiteArchiveInput,
    TaskFieldContainer,
    TaskInput,
    UnsignedMultiInput,
    UploadArtifactsInput,
)
from debusine.tasks.inputs.resolver import FieldResolver

__all__ = [
    "BuildArchitectureInput",
    "DebusineFQDNInput",
    "EnvironmentInput",
    "ExtraRepositoriesInput",
    "FieldResolver",
    "MultiInput",
    "MultiInputList",
    "OptionalEnvironmentInput",
    "OptionalSingleInput",
    "SingleInput",
    "SingleInputList",
    "Stage",
    "SuiteArchiveInput",
    "TaskFieldContainer",
    "TaskInput",
    "UnsignedMultiInput",
    "UploadArtifactsInput",
]
