# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import datetime as dt
import os
import sys

import django
from django.conf import settings

sys.path.insert(0, os.path.abspath('..'))
sys.path.insert(0, os.path.abspath('_ext'))

from docs.dynamic_data import DynamicDataDirective

# Run django's initialization code so that all modules can be loaded
# but disable configuration of logging as it can break sphinx-build
os.environ["DJANGO_SETTINGS_MODULE"] = "debusine.project.settings"
settings.LOGGING_CONFIG = None
django.setup()

# -- Project information -----------------------------------------------------

project = 'Debusine'
copyright = f'2019-{dt.date.today().year}, The Debusine Developers'
author = 'The Debusine Developers'

# The short X.Y version
version = ''
# The full version, including alpha/beta/rc tags
release = ''


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'debusine_sphinx',
    'sphinx.ext.autodoc',
    'sphinx.ext.extlinks',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
    'sphinx.ext.graphviz',
    'sphinx_copybutton',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = []

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# The name of the default domain.
primary_domain = 'debusine'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None

# Tweak copybutton so that it copies only commands and not prompts or outputs
copybutton_exclude = '.linenos, .gp'
copybutton_here_doc_delimiter = "END"
copybutton_line_continuation_character = "\\"
copybutton_prompt_text = r"\$ |# |root@debusine:~# "
copybutton_prompt_is_regexp = True
copybutton_only_copy_prompt_lines = True

# https://github.com/sphinx-doc/sphinx/issues/14223
suppress_warnings = ["ref.python"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    'logo_only': True,
    'style_nav_header_background': '#757575',
}
html_logo = 'images/debusine-logo-with-white-name.png'
html_favicon = 'images/favicon.png'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
html_sidebars = {
    '*': [
        'globaltoc.html',
        'relations.html',
        'sourcelink.html',
        'searchbox.html',
    ],
}

html_baseurl = "https://freexian-team.pages.debian.net/debusine/"


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'Debusinedoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        'Debusine.tex',
        'Debusine Documentation',
        'Raphaël Hertzog and other Debusine contributors',
        'manual',
    ),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, 'debusine', 'Debusine Documentation', [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        'Debusine',
        'Debusine Documentation',
        author,
        'Debusine',
        'One line description of project.',
        'Miscellaneous',
    ),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

intersphinx_mapping = {
    'python': ('https://docs.python.org/', None),
}

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# If true, `todolist` produces output without file path and line.
todo_link_only = True

# -- Options for autodoc extension -------------------------------------------
autodoc_default_options = {
    'members': True,
    'member-order': 'alphabetical',
    'special-members': '__init__',
    'undoc-members': True,
    'exclude-members': '__weakref__',
    'show-inheritance': True,
}

# -- Options for linkcheck extension -----------------------------------------
linkcheck_ignore = [
    r'https://matrix\.to',
    r'http://localhost',
    r'https://diataxis\.fr',
]

linkcheck_anchors_ignore = [
    r"^!",
    # GitLab comments
    r"^note_[0-9]+$",
    # Hetzner documentation
    r"^tag/servers/create_server$",
]

linkcheck_exclude_documents = [
    r"^reference/release-history$",
]

linkcheck_request_headers = {
    # These sites seem to block non-browser User-Agents, but work if we look
    # at least a bit like a browser.
    "https://repost.aws/": {"User-Agent": "Mozilla/5.0"},
    "https://docs.hetzner.com/": {"User-Agent": "Mozilla/5.0"},
    "https://www.hetzner.com/": {"User-Agent": "Mozilla/5.0"},
    # Report accurate User-Agent to various *.debian.org sites, instead of
    # the one defined by Sphinx (sphinx/util/requests.py).
    "https://salsa.debian.org/": {
        "User-Agent": "Debusine-Linkcheck (+https://debusine.debian.net/; sphinx-linkcheck)"
    },
    "https://bugs.debian.org/": {
        "User-Agent": "Debusine-Linkcheck (+https://debusine.debian.net/; sphinx-linkcheck)"
    },
    "https://wiki.debian.org/": {
        "User-Agent": "Debusine-Linkcheck (+https://debusine.debian.net/; sphinx-linkcheck)"
    },
}

# -- Options for extlinks extension ------------------------------------------
extlinks = {
    "issue": (
        "https://salsa.debian.org/freexian-team/debusine/-/work_items/%s",
        "#%s",
    ),
    "mr": (
        "https://salsa.debian.org/freexian-team/debusine/-/merge_requests/%s",
        "!%s",
    ),
}


def setup(app):
    app.add_directive("dynamic_data", DynamicDataDirective)
    app.add_css_file("css/debusine.css")
