# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine signon URLs Configuration."""

from django.urls import path

from debusine.server.signon.views import (
    BindIdentityView,
    OIDCAuthenticationCallbackView,
)

app_name = "signon"

urlpatterns = [
    path(
        "oidc_callback/<name>/",
        OIDCAuthenticationCallbackView.as_view(),
        name="oidc_callback",
    ),
    path(
        "bind_identity/<name>/",
        BindIdentityView.as_view(),
        name="bind_identity",
    ),
]
