# Generated by Django 4.2.28 on 2026-06-15 15:50

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0062_collectiontokengrant'),
    ]

    operations = [
        migrations.AddField(
            model_name='workflowtemplate',
            name='restricted',
            field=models.BooleanField(
                default=False,
                help_text='If true, only workspace owners can start this workflow, not contributors',
            ),
        ),
    ]
