# Generated by Django 4.2.20 on 2025-05-06 16:45

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0005_collectionitem_db_ci_suite_index_path_idx'),
    ]
    replaces = [
        ('db', '0147_fileinartifact_content_type'),
    ]

    operations = [
        migrations.AddField(
            model_name='fileinartifact',
            name='content_type',
            field=models.CharField(blank=True, null=True),
        ),
    ]
