{% extends base_template %}
{% load debusine %}
{% block content %}
    <div class="container">
        <div class="row">
            <div class="col-6">
                <form method="post">
                    {% csrf_token %}
                    {{ form.as_p }}
                    <input class="btn btn-primary btn-sm" type="submit" value="Save">
                    <a class="btn btn-secondary btn-sm"
                       href="{{ collection.get_absolute_url }}">Cancel</a>
                </form>
            </div>
        </div>
    </div>
{% endblock %}
