{% load debusine %}
{% if collection %}
    <a href="{{ collection.get_absolute_url }}">{{ value.name }}@{{ value.category }}</a>
{% else %}
    #{{ value.collection_id }} {{ value.name }}@{{ value.category }}
{% endif %}
