# Generated by Django 4.2.21 on 2025-11-03 12:10

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0024_workrequest_db_workrequest_workflow_has_root'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='artifact',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='artifactrelation',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='asset',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='assetusage',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='collection',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='collectionitem',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='file',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='group',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='groupauditlog',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='scope',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='token',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='user',
            options={
                'base_manager_name': 'objects',
                'verbose_name': 'user',
                'verbose_name_plural': 'users',
            },
        ),
        migrations.AlterModelOptions(
            name='worker',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='workerpool',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='workflowtemplate',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='workrequest',
            options={
                'base_manager_name': 'objects',
                'ordering': ['id'],
                'permissions': [
                    (
                        'manage_workrequest_priorities',
                        'Can set positive priority adjustments on work requests',
                    )
                ],
            },
        ),
        migrations.AlterModelOptions(
            name='workspace',
            options={'base_manager_name': 'objects'},
        ),
    ]
