/* This is a copy of the select2 default theme, edited to use color variables */
:root {
  --select2-background-color: var(--bs-tertiary-bg);
  --select2-border-color: var(--bs-border-color);
  --select2-choices-color: var(--bs-body-color);
  --select2-choices-background-color: var(--debusine-hover-layer);
  --select2-bg-item-current: var(--bs-primary);
  /* I could not see these colors being used, but I've abstracted them to
   * variables so if they show up we can target them here */
  --select2-unidentified1: #888;
  --select2-unidentified2: #eee;
  --select2-unidentified3: #333;
}
:root[data-bs-theme=light] {
  --select2-item-selected: oklab(from var(--select2-background-color) calc(l*0.9) a b / alpha);
}
:root[data-bs-theme=dark] {
  --select2-item-selected: oklab(from var(--select2-background-color) calc(l*1.5) a b / alpha);
}

.select2-container--debusine {
}
/* This was not themed in select2's default theme, we theme it here */
.select2-container--debusine .select2-dropdown {
  background-color: var(--select2-background-color);
  border-color: var(--select2-background-color);
}

.select2-container--debusine .select2-selection--single {
  background-color: var(--select2-background-color);
  border: 1px solid var(--select2-border-color);
  border-radius: 4px;
}
.select2-container--debusine .select2-selection--single .select2-selection__rendered {
  color: var(--select2-choices-color);
  line-height: 28px;
}
.select2-container--debusine .select2-selection--single .select2-selection__clear {
  cursor: pointer;
  float: right;
  font-weight: bold;
}
.select2-container--debusine .select2-selection--single .select2-selection__placeholder {
  color: var(--select2-choices-color);
}
.select2-container--debusine .select2-selection--single .select2-selection__arrow {
  height: 26px;
  position: absolute;
  top: 1px;
  right: 1px;
  width: 20px;
}
.select2-container--debusine .select2-selection--single .select2-selection__arrow b {
  border-color: var(--select2-unidentified1) transparent transparent transparent;
  border-style: solid;
  border-width: 5px 4px 0 4px;
  height: 0;
  left: 50%;
  margin-left: -4px;
  margin-top: -2px;
  position: absolute;
  top: 50%;
  width: 0;
}

.select2-container--debusine[dir="rtl"] .select2-selection--single .select2-selection__clear {
  float: left;
}

.select2-container--debusine[dir="rtl"] .select2-selection--single .select2-selection__arrow {
  left: 1px;
  right: auto;
}

.select2-container--debusine.select2-container--disabled .select2-selection--single {
  background-color: var(--select2-unidentified2);
  cursor: default;
}
.select2-container--debusine.select2-container--disabled .select2-selection--single .select2-selection__clear {
  display: none;
}

.select2-container--debusine.select2-container--open .select2-selection--single .select2-selection__arrow b {
  border-color: transparent transparent var(--select2-unidentified1) transparent;
  border-width: 0 4px 5px 4px;
}

.select2-container--debusine .select2-selection--multiple {
  background-color: var(--bs-tertiary-bg);
  border: 1px solid var(--select2-border-color);
  border-radius: 4px;
  cursor: text;
}
.select2-container--debusine .select2-selection--multiple .select2-selection__rendered {
  color: var(--select2-choices-color);
  box-sizing: border-box;
  list-style: none;
  margin: 0;
  padding: 0 5px;
  width: 100%;
}
.select2-container--debusine .select2-selection--multiple .select2-selection__rendered li {
  list-style: none;
}
.select2-container--debusine .select2-selection--multiple .select2-selection__clear {
  cursor: pointer;
  float: right;
  font-weight: bold;
  margin-top: 5px;
  margin-right: 10px;
  padding: 1px;
}
.select2-container--debusine .select2-selection--multiple .select2-selection__choice {
  background-color: var(--select2-choices-background-color);
  border: 1px solid var(--select2-border-color);
  border-radius: 4px;
  cursor: default;
  float: left;
  margin-right: 5px;
  margin-top: 5px;
  padding: 0 5px;
}
.select2-container--debusine .select2-selection--multiple .select2-selection__choice__remove {
  color: var(--select2-choices-color);
  cursor: pointer;
  display: inline-block;
  font-weight: bold;
  margin-right: 2px;
}
.select2-container--debusine .select2-selection--multiple .select2-selection__choice__remove:hover {
  color: var(--select2-unidentified3);
}

.select2-container--debusine[dir="rtl"] .select2-selection--multiple .select2-selection__choice, .select2-container--debusine[dir="rtl"] .select2-selection--multiple .select2-search--inline {
  float: right;
}

.select2-container--debusine[dir="rtl"] .select2-selection--multiple .select2-selection__choice {
  margin-left: 5px;
  margin-right: auto;
}

.select2-container--debusine[dir="rtl"] .select2-selection--multiple .select2-selection__choice__remove {
  margin-left: 2px;
  margin-right: auto;
}

.select2-container--debusine.select2-container--focus .select2-selection--multiple {
  border: solid black 1px;
  outline: 0;
}

.select2-container--debusine.select2-container--disabled .select2-selection--multiple {
  background-color: var(--select2-unidentified2);
  cursor: default;
}

.select2-container--debusine.select2-container--disabled .select2-selection__choice__remove {
  display: none;
}

.select2-container--debusine.select2-container--open.select2-container--above .select2-selection--single, .select2-container--debusine.select2-container--open.select2-container--above .select2-selection--multiple {
  border-top-left-radius: 0;
  border-top-right-radius: 0;
}

.select2-container--debusine.select2-container--open.select2-container--below .select2-selection--single, .select2-container--debusine.select2-container--open.select2-container--below .select2-selection--multiple {
  border-bottom-left-radius: 0;
  border-bottom-right-radius: 0;
}

.select2-container--debusine .select2-search--dropdown .select2-search__field {
  border: 1px solid var(--select2-border-color);
}

.select2-container--debusine .select2-search--inline .select2-search__field {
  background: transparent;
  border: none;
  outline: 0;
  box-shadow: none;
  -webkit-appearance: textfield;
}

.select2-container--debusine .select2-results > .select2-results__options {
  max-height: 200px;
  overflow-y: auto; }

.select2-container--debusine .select2-results__option[role=group] {
  padding: 0;
}

.select2-container--debusine .select2-results__option[aria-disabled=true] {
  color: var(--select2-choices-color);
}

.select2-container--debusine .select2-results__option[aria-selected=true] {
  background-color: var(--select2-item-selected);
}

.select2-container--debusine .select2-results__option .select2-results__option {
  padding-left: 1em;
}
.select2-container--debusine .select2-results__option .select2-results__option .select2-results__group {
  padding-left: 0;
}
.select2-container--debusine .select2-results__option .select2-results__option .select2-results__option {
  margin-left: -1em;
  padding-left: 2em;
}
.select2-container--debusine .select2-results__option .select2-results__option .select2-results__option .select2-results__option {
  margin-left: -2em;
  padding-left: 3em;
}
.select2-container--debusine .select2-results__option .select2-results__option .select2-results__option .select2-results__option .select2-results__option {
  margin-left: -3em;
  padding-left: 4em;
}
.select2-container--debusine .select2-results__option .select2-results__option .select2-results__option .select2-results__option .select2-results__option .select2-results__option {
  margin-left: -4em;
  padding-left: 5em;
}
.select2-container--debusine .select2-results__option .select2-results__option .select2-results__option .select2-results__option .select2-results__option .select2-results__option .select2-results__option {
  margin-left: -5em;
  padding-left: 6em;
}

.select2-container--debusine .select2-results__option--highlighted[aria-selected] {
  background-color: var(--select2-bg-item-current);
  color: white;
}

.select2-container--debusine .select2-results__group {
  cursor: default;
  display: block;
  padding: 6px;
}
