{% load artifacts %}
{% load debusine %}
<td class="text-nowrap">
    {# Artifact category #}
    <span class="bi bi-{{ artifact|artifact_icon_name }}"></span>
    {{ artifact|artifact_category_label|capfirst }}
</td>
<td>
    {# Artifact label (with link) #}
    <a class="text-decoration-none" href="{{ artifact.get_absolute_url }}">{{ artifact.get_label }}</a>
    {% if not artifact.complete %}(incomplete){% endif %}
</td>
<td>
    {# Action shortcuts: download, open #}
    <div class="btn-group btn-group-sm"
         role="group"
         aria-label="Artifact shortcuts">
        {% ui artifact as artifact_ui %}
        {% for shortcut in artifact_ui.ui_shortcuts %}
            {% widget shortcut %}
        {% endfor %}
    </div>
</td>
