# Generated by Django 4.2.23 on 2026-02-26 00:27

import django.db.models.fields.json
from django.db import migrations, models

import debusine.artifacts.models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0043_apt_authentication_assets'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='collectionitem',
            index=models.Index(
                models.F('parent_collection'),
                django.db.models.fields.json.KeyTextTransform(
                    'task_name', 'data'
                ),
                django.db.models.fields.json.KeyTextTransform(
                    'package', 'data'
                ),
                django.db.models.fields.json.KeyTextTransform(
                    'architecture', 'data'
                ),
                condition=models.Q(
                    (
                        'parent_category',
                        debusine.artifacts.models.CollectionCategory[
                            'QA_RESULTS'
                        ],
                    ),
                    ('removed_at__isnull', True),
                ),
                name='db_ci_qa_results_latest_idx',
            ),
        ),
        migrations.AddIndex(
            model_name='workrequest',
            index=models.Index(
                models.F('parent'),
                models.F('task_type'),
                models.F('task_name'),
                django.db.models.fields.json.KeyTextTransform(
                    'step', 'workflow_data_json'
                ),
                name='db_workrequest_wf_child_idx',
            ),
        ),
    ]
