{% extends base_template %}
{% load debusine %}
{% block content %}
    <p>Use the buttons below to trigger messages.</p>
    <form method="get">
        <button type="submit" class="btn btn-sm" name="type" value="debug">Debug</button>
        <button type="submit" class="btn btn-sm" name="type" value="info">Info</button>
        <button type="submit" class="btn btn-sm" name="type" value="success">Success</button>
        <button type="submit" class="btn btn-sm" name="type" value="warning">Warning</button>
        <button type="submit" class="btn btn-sm" name="type" value="error">Error</button>
    </form>
{% endblock %}
