# Generated by Django 4.2.28 on 2026-04-30 09:17

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0053_workrequesttoken'),
    ]

    operations = [
        migrations.AlterField(
            model_name='scoperole',
            name='role',
            field=models.CharField(
                choices=[('owner', 'Owner'), ('observer', 'Observer')]
            ),
        ),
    ]
