# Generated by Django 4.2.28 on 2026-04-28 10:25

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0051_collectionrelation'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collectionrelation',
            name='source',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='relations',
                to='db.collection',
            ),
        ),
        migrations.AlterField(
            model_name='collectionrelation',
            name='target',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='targeted_by',
                to='db.collection',
            ),
        ),
        migrations.AlterField(
            model_name='collectionrelation',
            name='type',
            field=models.CharField(
                choices=[
                    ('suite-forked-from', 'Forked from'),
                    ('suite-based-on', 'Based on'),
                    ('suite-requires', 'Requires'),
                    ('suite-targeting', 'Targeting'),
                    ('suite-default-qa-results', 'Default QA results'),
                ],
                max_length=32,
            ),
        ),
    ]
