# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""WorkflowTemplate task_data gets renamed, preparing for runtime_parameters."""

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0031_rename_more_arch_all_build_architecture_and_suite'),
    ]

    operations = [
        migrations.RenameField(
            model_name='workflowtemplate',
            old_name='task_data',
            new_name='static_parameters',
        ),
    ]
