# Generated by Django 4.2.21 on 2025-10-30 17:55

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0023_workrequest_workflow_root_data'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='workrequest',
            constraint=models.CheckConstraint(
                check=models.Q(
                    ('parent__isnull', True),
                    ('workflow_root__isnull', False),
                    _connector='OR',
                ),
                name='db_workrequest_workflow_has_root',
            ),
        ),
    ]
