# Generated by Django 4.2.21 on 2025-09-03 13:05

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0015_alter_workspace_expiration_delay'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='artifact',
            index=models.Index(
                models.F('workspace'),
                models.F('category'),
                include=('id',),
                name='db_artifact_workspace_category',
            ),
        ),
    ]
