{% load debusine %}
<table class="table table-sm table-hover table-striped"
       id="worker_pool-list">
    {% widget paginator.thead %}
    <tbody>
        {% for wp in paginator.page_obj.object_list %}
            <tr>
                <td>
                    <a href="{{ wp.get_absolute_url }}">{{ wp.name }}</a>
                </td>
                <td>{{ wp.enabled|yesno }}</td>
                <td>{{ wp.instance_wide|yesno }}</td>
                <td>{{ wp.ephemeral|yesno }}</td>
                <td>{{ wp.workers }}</td>
            </tr>
        {% empty %}
            <tr>
                <td colspan="{{ table.columns|length }}">There are no rows to show matching the current filter</td>
            </tr>
        {% endfor %}
    </tbody>
    {% widget paginator.tfoot %}
</table>
