{% if result == "success" %}
    <span class="badge text-bg-success">Success</span>
{% elif result == "failure" %}
    <span class="badge text-bg-warning">Failure</span>
{% elif result == "error" %}
    <span class="badge text-bg-danger">Error</span>
{% elif result == "skipped" %}
    <span class="badge text-bg-secondary">Skipped</span>
{% endif %}
