{% extends "web/_file_base.html" %}
{% block formatted_content %}
    Binary file {{ file_in_artifact.path }} cannot be displayed.
    You can <a href="{{ file_in_artifact.get_absolute_url_raw }}">view it raw</a>
    or <a href="{{ file_in_artifact.get_absolute_url_download }}">download</a> it instead.
{% endblock %}
