{% extends base_template %}
{% load debusine %}
{% block breadcrumbs %}{% endblock %}
{% block content %}
    {% block introduction %}
        <h1 class="mt-4 mb-4">Welcome to Debusine!</h1>
        <div class="row">
            <div class="col-10 col-xxl-8">
                <p>
                    Debusine manages scheduling and distribution of Debian-related tasks to a
                    distributed network of machines.
                </p>
                <p>
                    It is under active development by <a href="https://www.freexian.com/">Freexian</a>
                    with the intention of providing easy access to a range of packaging tools
                    and distribution-wide experiments.
                </p>
                {% if user.is_authenticated %}
                    <p>
                        You are authenticated as <b>{{ user.username }}</b>.
                    </p>
                {% endif %}
                <p>
                    If it's your first time exploring Debusine, you can start from
                    <a href="https://freexian-team.pages.debian.net/debusine/explanation/introduction.html#what-is-debusine">What
                    is Debusine</a>
                    for an introduction, or browse the index to the whole
                    <a href="https://freexian-team.pages.debian.net/debusine/index.html">Debusine documentation</a>.
                    You may want to
                    <a href="https://freexian-team.pages.debian.net/debusine/howtos/set-up-debusine-client.html">set up debusine-client</a>
                    if you have not already done so.
                </p>
            </div>
        </div>
    {% endblock %}
    {% if user.is_authenticated %}
        <div class="row">
            <div class="col-10 col-xxl-8">
                <h2 class="my-4">Workflows</h2>
                <p>
                    <a href="https://freexian-team.pages.debian.net/debusine/explanation/concepts.html#workflows">Workflows</a>
                    are how Debusine orchestrates and manages sets of
                    <a href="https://freexian-team.pages.debian.net/debusine/explanation/concepts.html#tasks">tasks</a>
                    to perform the operations requested by users.
                    Here you can see a list of your current and past workflows processed by Debusine.
                </p>
            </div>
        </div>
        <div class="row">
            <div class="col-12 col-xxl-6">
                <h3 class="my-4">Your running workflows</h3>
                {% widget workflows_current %}
            </div>
            <div class="col-12 col-xxl-6">
                <h3 class="my-4">Your completed workflows</h3>
                {% widget workflows_completed %}
            </div>
        </div>
    {% endif %}
    <h2 class="my-4">Scopes and workspaces</h2>
    <div class="row">
        <div class="col-10 col-xxl-8">
            <p>
                Everything in Debusine happens inside
                <a href="https://freexian-team.pages.debian.net/debusine/explanation/concepts.html#workspaces">workspaces</a>.
                Workspaces exist inside <a href="https://freexian-team.pages.debian.net/debusine/explanation/concepts.html#scopes">scopes</a>
                which act as administrative divisions.
            </p>
        </div>
    </div>
    {% widget workspaces %}
{% endblock %}
