{% load debusine %}
<h2>Output artifacts</h2>
{% if not built_artifacts_table %}
    <p id="no-output-artifacts">No artifacts produced.</p>
{% else %}
    {% widget built_artifacts_table %}
{% endif %}
<h2>Output data</h2>
{% if work_request_output_data %}
    {% include "web/_dict_to_ul.html" with dict=work_request_output_data only %}
{% else %}
    <p id="output-no-output-data">No output data.</p>
{% endif %}
