# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Debusine command line interface, archive management commands.

These are "porcelain": they implement opinionated operations on
:collection:`debian:archive` collections and their contents, rather than
basic "plumbing" operations.
"""

# Commands need to be imported in order to be available to cli.py.
# They are registered via Command.__init_subclass__.

import debusine.client.commands.archives.suites  # noqa: F401
import debusine.client.commands.archives.tokens  # noqa: F401
