{% load debusine %}
<div class="row">
    {% if autopkgtest_artifact_error %}
        <div id="error-artifact" class="alert alert-warning">Cannot display autopkgtest: {{ autopkgtest_artifact_error }}</div>
    {% else %}
        <div class="col-md-2 list-group">
            {% for section in sections %}
                <div class="list-group mb-3">
                    <div class="list-group-item list-group-item-{{ section.color }}">
                        <strong>{{ section.text }}</strong>
                    </div>
                    {% for subsection in section.subsections %}
                        {% if subsection.include_in_table_of_contents %}
                            <a class="list-group-item {% if subsection.color %}list-group-item-{{ subsection.color }}{% endif %}"
                               {% if title %}title="{{ subsection.tooltip }}"{% endif %}
                               href="#L{{ subsection.line }}-section">{{ subsection.text }}</a>
                        {% endif %}
                    {% endfor %}
                </div>
            {% endfor %}
        </div>
        {% if log_widget %}
            <div class="col-md-10">{% widget log_widget %}</div>
        {% endif %}
    {% endif %}
</div>
