{% load debusine %}
<ul>
    {% for key, value in dict.items %}
        <li>
            <b>{{ key }}:</b>
            {% if value is None %}
                -
            {% elif value|is_dict_like %}
                {% include "web/_dict_to_ul.html" with dict=value only %}
            {% elif value|is_list_like %}
                {% include "web/_list_to_ul.html" with list=value only %}
            {% else %}
                {{ value }}
            {% endif %}
        </li>
    {% endfor %}
</ul>
