# Generated by Django 4.2.23 on 2025-12-12 13:57

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0036_rename_create_child_workspace'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collection',
            name='full_history_retention_period',
            field=models.DurationField(
                blank=True,
                help_text='Period after removing an item from the collection during which the related artifact is not allowed to expire',
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name='collection',
            name='metadata_only_retention_period',
            field=models.DurationField(
                blank=True,
                help_text='Period after removing an item from the collection during which its metadata is kept in the history of the collection',
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name='collection',
            name='retains_artifacts',
            field=models.CharField(
                choices=[
                    ('never', 'Never'),
                    ('workflow', 'While workflow is running'),
                    ('always', 'Always'),
                ],
                default='always',
                help_text='Controls if artifacts in this collection should be retained rather than being allowed to expire',
                max_length=8,
            ),
        ),
    ]
