<button type="button"
        class="btn btn-link p-0 ms-1 text-body"
        aria-label="Show lookup"
        title="Lookup"
        aria-haspopup="dialog"
        aria-controls="lookup-modal-{{ id_suffix }}"
        data-bs-toggle="modal"
        data-bs-target="#lookup-modal-{{ id_suffix }}">
    <i class="bi bi-search"></i>
</button>
<div class="modal"
     id="lookup-modal-{{ id_suffix }}"
     tabindex="-1"
     aria-labelledby="lookup-modal-{{ id_suffix }}-label"
     aria-hidden="true">
    <div class="modal-dialog modal-lg modal-dialog-scrollable">
        <div class="modal-content">
            <div class="modal-header">
                <h5 class="modal-title" id="lookup-modal-{{ id_suffix }}-label">
                    Lookup: <code>{{ label }}</code>
                </h5>
                <button type="button"
                        class="btn-close"
                        data-bs-dismiss="modal"
                        aria-label="Close"></button>
            </div>
            <div class="modal-body">
                <div class="card">{% include "web/_card-body.html" with data=lookup_display only %}</div>
            </div>
            <div class="modal-footer">
                <button type="button" class="btn btn-secondary" data-bs-dismiss="modal">Close</button>
            </div>
        </div>
    </div>
</div>
