# Generated by Django 3.2.19 on 2024-07-12 13:26

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('signing', '0002_key_discriminated_union'),
    ]

    operations = [
        migrations.AlterField(
            model_name='auditlog',
            name='created_by_work_request_id',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='auditlog',
            name='event',
            field=models.CharField(
                choices=[
                    ('generate', 'Generate'),
                    ('sign', 'Sign'),
                    ('register', 'Register an externally-generated key'),
                ],
                max_length=8,
            ),
        ),
    ]
