# Generated by Django 4.2.23 on 2025-12-12 12:14

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0039_alter_worker_worker_type_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='workrequest',
            name='pending_at',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]
