# Generated by Django 4.2.21 on 2025-09-16 11:22

from django.db import migrations, models

import debusine.assets.models


class Migration(migrations.Migration):
    dependencies = [
        ('signing', '0003_register_pkcs11_static_key'),
    ]

    operations = [
        migrations.AlterField(
            model_name='auditlog',
            name='purpose',
            field=models.CharField(
                choices=[
                    (
                        debusine.assets.models.KeyPurpose['UEFI'],
                        debusine.assets.models.KeyPurpose['UEFI'],
                    ),
                    (
                        debusine.assets.models.KeyPurpose['OPENPGP_UPLOAD'],
                        debusine.assets.models.KeyPurpose['OPENPGP_UPLOAD'],
                    ),
                    (
                        debusine.assets.models.KeyPurpose['OPENPGP_REPOSITORY'],
                        debusine.assets.models.KeyPurpose['OPENPGP_REPOSITORY'],
                    ),
                ],
                max_length=32,
            ),
        ),
        migrations.AlterField(
            model_name='key',
            name='purpose',
            field=models.CharField(
                choices=[
                    (
                        debusine.assets.models.KeyPurpose['UEFI'],
                        debusine.assets.models.KeyPurpose['UEFI'],
                    ),
                    (
                        debusine.assets.models.KeyPurpose['OPENPGP_UPLOAD'],
                        debusine.assets.models.KeyPurpose['OPENPGP_UPLOAD'],
                    ),
                    (
                        debusine.assets.models.KeyPurpose['OPENPGP_REPOSITORY'],
                        debusine.assets.models.KeyPurpose['OPENPGP_REPOSITORY'],
                    ),
                ],
                max_length=32,
            ),
        ),
    ]
