{% load static %}
{% if provider.icon %}
    <img style="vertical-align: text-top;
                height: 1em"
         class="mr-2"
         alt="{{ provider.label }} icon"
         src="{% static provider.icon %}">
{% else %}
    <i class="mr-2 bi bi-box-arrow-in-right"></i>
{% endif %}
