{% load debusine %}
<table class="table table-hover table-striped task-input-fields"
       id="{{ widget.element_id }}">
    <thead>
        <tr>
            <th>Name</th>
            <th>Value</th>
        </tr>
    </thead>
    <tbody>
        {% for field in widget.fields %}
            {% widget field %}
        {% endfor %}
    </tbody>
</table>
