{% extends base_template %}
{% block content %}
    <form method="post">
        {% csrf_token %}
        {{ form.as_p }}
        <input class="btn btn-primary btn-sm" type="submit" value="{{ action }}">
    </form>
{% endblock %}
