# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Views for debusine."""

from debusine.web.views.autopkgtest import AutopkgtestViewWorkRequestPlugin
from debusine.web.views.debdiff import DebDiffViewArtifactPlugin
from debusine.web.views.debian_pipeline import (
    DebianPipelineViewWorkRequestPlugin,
)
from debusine.web.views.lintian import LintianViewWorkRequestPlugin
from debusine.web.views.qa import QAViewWorkRequestPlugin
from debusine.web.views.views import HomepageView

__all__ = [
    "AutopkgtestViewWorkRequestPlugin",
    "DebDiffViewArtifactPlugin",
    "DebianPipelineViewWorkRequestPlugin",
    "HomepageView",
    "LintianViewWorkRequestPlugin",
    "QAViewWorkRequestPlugin",
]
