{% extends "web/_file_base.html" %}
{% block formatted_content %}
    <pre class="formatted-content"
         style="--file-lineno-width: {{ file_lineno_width }}em">{{ file_content }}</pre>
{% endblock %}
