{% if result == "success" %}
    <span title="Success" class="badge text-bg-success">S</span>
{% elif result == "failure" %}
    <span title="Failure" class="badge text-bg-warning">F</span>
{% elif result == "error" %}
    <span title="Error" class="badge text-bg-danger">E</span>
{% elif result == "skipped" %}
    <span title="Skipped" class="badge text-bg-secondary">s</span>
{% else %}
    {{ result }}
{% endif %}
