# Generated by Django 4.2.20 on 2025-06-18 13:47

from django.db import migrations, models

import debusine.artifacts.models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0013_qa_suite'),
    ]
    replaces = [
        ('db', '0155_alter_workrequest_result'),
    ]

    operations = [
        migrations.AlterField(
            model_name='workrequest',
            name='result',
            field=models.CharField(
                choices=[
                    (debusine.artifacts.models.WorkRequestResults['NONE'], ''),
                    (
                        debusine.artifacts.models.WorkRequestResults['SUCCESS'],
                        'Success',
                    ),
                    (
                        debusine.artifacts.models.WorkRequestResults['FAILURE'],
                        'Failure',
                    ),
                    (
                        debusine.artifacts.models.WorkRequestResults['ERROR'],
                        'Error',
                    ),
                    (
                        debusine.artifacts.models.WorkRequestResults['SKIPPED'],
                        'Skipped',
                    ),
                ],
                default=debusine.artifacts.models.WorkRequestResults['NONE'],
                editable=False,
                max_length=7,
            ),
        ),
    ]
