# Generated by Django 5.2.13 on 2026-05-06 16:14

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
        ('db', '0057_filestorejournal'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='user',
            constraint=models.CheckConstraint(
                check=models.Q(('username__regex', '^\\d+$'), _negated=True),
                name='db_user_non_integer_username',
            ),
        ),
    ]
