{% extends base_template %}
{% load debusine %}
{% block content %}
    <div id="archive-suites">
        {% if suites %}
            <p>This archive contains the following suites:</p>
            <ul>
                {% for suite in suites %}
                    <li>
                        <p>
                            <a href="{% url 'suite-root' scope=scope workspace=workspace suite=suite.name %}">{{ suite.name }}</a>: {{ suite.description|default_if_none:"" }}
                        </p>
                        <pre><code>{{ suite.as_deb822_source }}</code></pre>
                    </li>
                {% endfor %}
            </ul>
        {% else %}
            <p>This archive is empty.</p>
        {% endif %}
    </div>
{% endblock %}
