{% load debusine %}
<table class="table table-sm table-hover table-striped"
       id="{{ table.prefix }}work_request-list">
    {% widget paginator.thead %}
    <tbody>
        {% for work_request in paginator.page_obj.object_list %}
            {% ui work_request as work_request_ui %}
            <tr>
                <td>
                    <a href="{{ work_request.get_absolute_url }}">{{ work_request.id }}</a>
                </td>
                <td title="{{ work_request.created_at }}">{{ work_request.created_at|timesince }}</td>
                <td>{{ work_request.task_type }}</td>
                <td title="{{ work_request.task_name }}">{{ work_request_ui.label }}</td>
                <td>
                    {% include "web/_work_request-status.html" with status=work_request.status aborted_by=work_request.aborted_by only %}
                </td>
                <td>{% include "web/_work_request-result.html" with result=work_request.result only %}</td>
            </tr>
        {% empty %}
            <tr>
                <td colspan="{{ table.columns|length }}">There are no rows to show matching the current filter</td>
            </tr>
        {% endfor %}
    </tbody>
    {% widget paginator.tfoot %}
</table>
