# Generated by Django 4.2.21 on 2025-09-17 12:29

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0017_assetusage_db_assetusage_unique_asset_workspace'),
    ]

    operations = [
        migrations.AlterField(
            model_name='assetusagerole',
            name='role',
            field=models.CharField(
                choices=[
                    ('signer', 'Signer'),
                    ('repository_signer', 'Repository signer'),
                ],
                max_length=32,
            ),
        ),
    ]
