{% extends base_template %}
{% load debusine %}
{% block content %}
    {% ui workflow_template as workflow_template_ui %}
    <table class="table table-hover table-striped" id="workflow-information">
        <colgroup>
            <col style="width: 1%">
            <col>
        </colgroup>
        <tr>
            <th class="text-nowrap">Template name:</th>
            <td>{{ workflow_template.name }}</td>
        </tr>
        <tr>
            <th class="text-nowrap">Task name:</th>
            <td>
                <a href="{{ workflow_template_ui.get_help_url }}">{{ workflow_template.task_name }}</a>
            </td>
        </tr>
        <tr>
            <th class="text-nowrap">Priority:</th>
            <td>{{ workflow_template.priority }}</td>
        </tr>
        <tr>
            <th colspan="2" class="text-nowrap" style="border-bottom: none;">Static parameters:</th>
        </tr>
        <tr>
            <td colspan="2">
                <div id="task-data" class="card">
                    <div class="card-body text-white bg-dark rounded">{{ workflow_template.static_parameters|as_yaml }}</div>
                </div>
            </td>
        </tr>
        <tr>
            <th colspan="2" class="text-nowrap" style="border-bottom: none;">Runtime parameters:</th>
        </tr>
        <tr>
            <td colspan="2">
                <div id="runtime-parameters" class="card">
                    <div class="card-body text-white bg-dark rounded">{{ workflow_template.runtime_parameters|as_yaml }}</div>
                </div>
            </td>
        </tr>
    </table>
{% endblock %}
