{% load debusine %}
<dl>
    <dt>Task</dt>
    <dd>
        Type: <b>{{ work_request.task_type }}</b>.
        Name: <b>{{ work_request.task_name }}</b>.
        {% with workflow_root=work_request.workflow_root %}
            {% if workflow_root %}
                {% with collection=workflow_root.internal_collection %}
                    Internal collection:
                    <b><a id="internal-collection" href="{{ collection.get_absolute_url }}">{{ collection.name }}</a></b>.
                {% endwith %}
            {% endif %}
        {% endwith %}
    </dd>
    <dt>Task data</dt>
    <dd>
        {% include "web/_card.html" with id="card-task_data" data=task_data only %}
    </dd>
    {% if show_task_data_configured %}
        <dt>Task data after applying task configuration</dt>
        <dd>
            {% include "web/_card.html" with id="card-configured_task_data" data=task_data_configured only %}
        </dd>
        <dt>Scheduler tags</dt>
        <dd>
            {% include "web/_card.html" with id="card-scheduler-tags" data=scheduler_tags only %}
        </dd>
    {% else %}
        <dt>Task data after applying task configuration / scheduler tags</dt>
        <dd>
            <p id="task_data_not_configured">Task has not yet been configured.</p>
        </dd>
    {% endif %}
    {% if input_fields_internal or task_input_fields_raw %}
        <dt>Input fields</dt>
        <dd>
            {% if input_fields_internal %}
                {% widget input_fields_internal %}
            {% endif %}
            {% if task_input_fields_raw %}
                {% include "web/_card.html" with id="card-task-input-fields-raw" data=task_input_fields only %}
            {% endif %}
        </dd>
    {% endif %}
    <dt>Dynamic data</dt>
    <dd>
        {% if show_task_data_dynamic %}
            {% include "web/_card.html" with id="card-dynamic_data" data=task_data_dynamic only %}
        {% else %}
            <p id="dynamic_data_not_available">Dynamic task data has not yet been computed.</p>
        {% endif %}
    </dd>
    <dt>Workflow data</dt>
    <dd>
        {% include "web/_card.html" with id="card-workflow_data" data=work_request_workflow_data_json only %}
    </dd>
    <dt>Event reactions</dt>
    <dd>
        {% include "web/_card.html" with id="card-event_reactions" data=work_request_event_reactions_json only %}
    </dd>
    <dt>Output data</dt>
    <dd>
        {% if work_request_output_data %}
            {% include "web/_card.html" with id="card-output_data" data=work_request_output_data_json only %}
        {% else %}
            <p id="internals-no-output-data">No output data.</p>
        {% endif %}
    </dd>
</dl>
