{% extends "web/_base.html" %}
{% load static %}
{% load debusine %}
{% load archives %}
{% block navbar %}
    <nav class="navbar navbar-expand-lg bg-body-tertiary">
        <div class="container-fluid">
            <ul id="navbar-homepage" class="navbar-nav me-auto mb-2 mb-lg-0">
                <li class="nav-item">
                    <a id="navbar-homepage-link"
                       class="navbar-brand btn"
                       href="{% debusine_url 'workspaces:detail' wname=workspace %}">
                        <img src="{% static 'web/icons/debusine.png' %}"
                             height="25"
                             alt="Homepage"
                             class="d-inline-block align-text-top">
                        {% block navbar_brand_name %}Debusine{% endblock %}
                    </a>
                </li>
            </ul>
            <button class="navbar-toggler"
                    type="button"
                    data-bs-toggle="collapse"
                    data-bs-target="#navbarSupportedContent"
                    aria-controls="navbarSupportedContent"
                    aria-expanded="false"
                    aria-label="Toggle navigation">
                <span class="navbar-toggler-icon"></span>
            </button>
            <div class="collapse navbar-collapse" id="navbarSupportedContent">
                <ul id="navbar-left" class="navbar-nav me-auto mb-2 mb-lg-0">
                    {% block navbar_left %}
                        <li class="nav-item">
                            <a class="nav-link btn btn-debusine"
                               href="{% url 'archive-root' scope=scope workspace=workspace %}">{{ workspace }}</a>
                        </li>
                    {% endblock %}
                </ul>
                <ul id="navbar-right" class="navbar-nav mb-2 mb-lg-0">
                    {% block navbar_right %}{% endblock %}
                </ul>
            </div>
        </div>
    </nav>
{% endblock %}
{% block footer %}
    {% include "web/footer_archives.html" %}
{% endblock %}
