# Generated by Django 4.2.21 on 2025-10-30 15:53

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0021_workrequest_completed_by'),
    ]

    operations = [
        migrations.AddField(
            model_name='workrequest',
            name='workflow_root',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name='descendants',
                to='db.workrequest',
            ),
        ),
    ]
