{% extends "web/_base_scoped.html" %}
{% load debusine %}
{% block main-layout %}
    <div class="row">
        <div class="col-8">
            {% block pagetitle %}{{ place.as_page_title }}{% endblock %}
            {% block content %}{% endblock %}
        </div>
        <div class="col-4">
            <div class="btn-group"
                 role="toolbar"
                 aria-label="Shortcuts"
                 id="right-sidebar-shortcuts">
                {% block ui-shortcuts %}
                    {% ui object as object_ui %}
                    {% for shortcut in object_ui.ui_shortcuts %}
                        {% widget shortcut %}
                    {% endfor %}
                {% endblock %}
            </div>
            {% block sidebar %}
                {% if sidebar_items %}
                    <div id="workrequest-sidebar" class="mt-2 list-group">
                        {% for item in sidebar_items %}
                            {% widget item %}
                        {% endfor %}
                    </div>
                {% endif %}
            {% endblock %}
            {% block content-right %}{% endblock %}
        </div>
    </div>
{% endblock %}
