# Generated by Django 4.2.21 on 2025-08-13 14:39

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0014_alter_workrequest_result'),
    ]

    operations = [
        migrations.AlterField(
            model_name='workspace',
            name='expiration_delay',
            field=models.DurationField(
                blank=True,
                help_text='if set, time since the last task completion time after which the workspace can be deleted',
                null=True,
            ),
        ),
    ]
