{% load debusine %}
<td class="dropend"
    {% if rows|length > 1 %}rowspan="{{ rows|length }}"{% endif %}>
    <div class="name w-100"
         role="button"
         data-bs-toggle="dropdown"
         data-bs-auto-close="outside"
         aria-expanded="false">
        <span>{{ widget.verbose_name }}</span>
        <span class="details bi bi-search dropdown-toggle-split"></span>
    </div>
    <div class="dropdown-menu p-4">{{ widget.render_details }}</div>
</td>
