# Generated by Django 4.2.20 on 2025-04-29 10:39

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0001_squashed_0_11_0'),
    ]
    replaces = [
        ('db', '0143_workrequest_aborted_by'),
    ]

    operations = [
        migrations.AddField(
            model_name='workrequest',
            name='aborted_by',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name='aborted_work_requests',
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]
