# Generated by Django 4.2.28 on 2026-05-24 07:44

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0058_non_numeric_usernames'),
    ]

    operations = [
        migrations.AddField(
            model_name='workflowtemplate',
            name='extra_groups',
            field=models.ManyToManyField(
                help_text='Extra groups granted when running a workflow',
                related_name='granted_by_workflow_template',
                to='db.group',
            ),
        ),
    ]
