# Generated by Django 4.2.20 on 2025-05-05 19:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0002_workrequest_aborted_by'),
    ]
    replaces = [
        ('db', '0144_add_workspace_role_viewer'),
    ]

    operations = [
        migrations.AlterField(
            model_name='workspacerole',
            name='role',
            field=models.CharField(
                choices=[
                    ('owner', 'Owner'),
                    ('contributor', 'Contributor'),
                    ('viewer', 'Viewer'),
                ],
                max_length=16,
            ),
        ),
    ]
