{% block footer_copyright_row %}
    <div id="footer-copyright" class="row justify-content-md-center">
        <div class="col-auto">
            © 2019-2026 <a href="https://salsa.debian.org/freexian-team/debusine/-/blob/devel/AUTHORS">Debusine Developers</a>
        </div>
    </div>
{% endblock %}
{% block footer_utils %}
    <div id="footer-utils" class="row justify-content-md-center">
        {% block footer_status %}
            <div id="footer-status" class="col-auto">
                Status:
                <a href="{% url "worker-pools:list" %}">Worker pools</a>
                • <a href="{% url "workers:list" %}">Workers</a>
                • <a href="{% url "task-status" %}">Task status</a>
            </div>
        {% endblock %}
        {% block footer_theme %}
            <form id="footer-theme"
                  class="col-auto"
                  method="post"
                  action="{% url "theme-selection" %}">
                {% csrf_token %}
                <input type="hidden" name="next" value="{{ request.get_full_path }}">
                Theme:
                <button type="submit" name="theme" value="dark">dark</button>
                •
                <button type="submit" name="theme" value="light">light</button>
                •
                <button type="submit" name="theme" value="auto">auto</button>
            </form>
        {% endblock %}
    </div>
{% endblock %}
{% block footer_links %}
    <div id="footer-software" class="row justify-content-md-center">
        <div class="col-auto">
            Debusine software project:
            <a href="https://freexian-team.pages.debian.net/debusine/"
               target="_blank">Documentation</a> •
            <a href="https://salsa.debian.org/freexian-team/debusine/-/issues"
               target="_blank">Bugs</a> •
            <a href="https://salsa.debian.org/freexian-team/debusine/"
               target="_blank">Code</a> •
            <a href="https://freexian-team.pages.debian.net/debusine/howtos/contribute.html"
               target="_blank">Contributing</a>
            {% if debusine_version %}• {{ debusine_version }}{% endif %}
        </div>
    </div>
{% endblock %}
{% block footer_extra %}
    <div id="footer-extra" class="row justify-content-md-center">
        <div class="col-auto">
            Commercial support: <a href="https://www.freexian.com/services/debusine/" target="_blank">Freexian</a>
        </div>
    </div>
{% endblock %}
