{% load artifacts %}
{% if data %}
    {% if data.source_artifact_original_id and data.source_artifact_new_id %}
        <p id="description-source_artifacts">
            Compare source package {% artifact_link data.source_artifact_original_id %}
            with {% artifact_link data.source_artifact_new_id %} using debdiff
        </p>
    {% elif data.source_package_names %}
        <p id="description-binary_artifacts">
            Compare {{ data.binary_artifacts_architecture }}
            {{ data.binary_artifacts_new_count|pluralize:"binary,binaries" }} from
            source package{{ data.source_package_names|length|pluralize }}
            {{ data.source_package_names|join:", " }}
            using debdiff.
        </p>
    {% else %}
        <p id="general-description">Debdiff for {{ data.parameter_summary }}</p>
    {% endif %}
    <ul>
        <li id="description-executed_in_environment">
            Executed in environment
            {% artifact_link data.environment_id %}
        </li>
        {% if data.extra_arguments %}
            <li id="description-extra_arguments">
                Extra arguments:
                {% include "web/_list_to_code.html" with values=data.extra_arguments only %}
            </li>
        {% endif %}
    </ul>
{% endif %}
