# Generated by Django 4.2.20 on 2025-06-06 11:49

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0011_collectionitem_db_ci_archive_index_path_idx'),
    ]
    replaces = [
        ('db', '0153_alter_collectionitem_created_by_workflow_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collectionitem',
            name='created_by_workflow',
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name='workflow_created_%(class)s',
                to='db.workrequest',
            ),
        ),
        migrations.AlterField(
            model_name='collectionitem',
            name='removed_by_workflow',
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name='workflow_removed_%(class)s',
                to='db.workrequest',
            ),
        ),
    ]
