# Generated by Django 4.2.20 on 2025-05-25 19:56

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0007_alter_collectionitem_created_at_and_more'),
    ]
    replaces = [
        ('db', '0149_collectionitem_db_ci_name_idx'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='collectionitem',
            index=models.Index(models.F('name'), name='db_ci_name_idx'),
        ),
    ]
