# Generated by Django 4.2.28 on 2026-03-24 16:29

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0046_copy_build_arch_to_tag'),
    ]

    operations = [
        migrations.AlterField(
            model_name='worker',
            name='activation_token',
            field=models.OneToOneField(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name='activating_worker',
                to='db.token',
            ),
        ),
    ]
