# Generated by Django 4.2.23 on 2026-02-04 17:30

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0041_add_group_workspace_namespace'),
    ]

    operations = [
        migrations.AddField(
            model_name='workrequest',
            name='scheduler_tags_provided',
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=255),
                blank=True,
                default=list,
                help_text='final set of task-provided tags used for scheduling',
                size=None,
            ),
        ),
        migrations.AddField(
            model_name='workrequest',
            name='scheduler_tags_required',
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=255),
                blank=True,
                default=list,
                help_text='final set of task-required tags used for scheduling',
                size=None,
            ),
        ),
    ]
