<!DOCTYPE html>
<html lang="en">
    <head>
        <title>403 Forbidden</title>
    </head>
    <body>
        <h1>403 Forbidden</h1>
        {% if exception %}<p>{{ exception }}</p>{% endif %}
    </body>
</html>
