# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""URLs related to collections in workspaces."""

from django.urls import path

import debusine.web.views.collection_relations as views

app_name = "collection_relations"

urlpatterns = [
    path(
        "",
        views.CollectionRelationListView.as_view(),
        name="list",
    ),
    path(
        "<int:pk>/delete/",
        views.CollectionRelationDeleteView.as_view(),
        name="delete",
    ),
    path(
        "create/",
        views.CollectionRelationCreateView.as_view(),
        name="create",
    ),
    path(
        "reorder-requires/",
        views.CollectionRelationReorderRequiresView.as_view(),
        name="reorder_requires",
    ),
]
