# Generated by Django 4.2.28 on 2026-03-03 14:02

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0044_collectionitem_db_ci_qa_results_latest_idx_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='workrequest',
            name='unblock_strategy',
            field=models.CharField(
                choices=[
                    ('deps', 'Dependencies have completed'),
                    ('manual', 'Manually unblocked'),
                    (
                        'incremental',
                        'May be unblocked at any time to make incremental progress',
                    ),
                ],
                default='deps',
                editable=False,
                max_length=11,
            ),
        ),
    ]
