<i class="bi bi-question-circle fs-6 align-middle"
   tabindex="0"
   role="button"
   data-bs-toggle="popover"
   data-bs-trigger="focus"
   data-bs-html="true"
   title="{{ title }}"
   data-bs-content="{{ summary|safe }}<br><a href='{{ link }}'>Learn more</a>">
</i>
