# Generated by Django 4.2.21 on 2025-09-16 10:17

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0016_artifact_db_artifact_workspace_category'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='assetusage',
            constraint=models.UniqueConstraint(
                fields=('asset', 'workspace'),
                name='db_assetusage_unique_asset_workspace',
            ),
        ),
    ]
