# Generated by Django 5.2.13 on 2026-05-05 16:39

from django.db import migrations, models

import debusine.db.models.auth


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0055_system_apt_credentials'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='token',
            name='db_token_valid_type',
        ),
        migrations.AddConstraint(
            model_name='token',
            constraint=models.CheckConstraint(
                check=models.Q(
                    (
                        'token_type__in',
                        [
                            debusine.db.models.auth._TokenTypes.USER,
                            debusine.db.models.auth._TokenTypes.WORK_REQUEST,
                            debusine.db.models.auth._TokenTypes.WORKER,
                            debusine.db.models.auth._TokenTypes.WORKER_ACTIVATION,
                        ],
                    )
                ),
                name='db_token_valid_type',
            ),
        ),
    ]
