{% load debusine %}
<h2>
    Dependencies
    {% help "dependencies" %}
</h2>
{% if work_request_dependencies %}
    <p>Work requests that must be completed before this one can run.</p>
    {% widget work_request_dependencies %}
{% else %}
    <p>This work request does not depend on any other one.</p>
{% endif %}
<h2>
    Reverse dependencies
    {% help "reverse_dependencies" %}
</h2>
{% if work_request_reverse_dependencies %}
    <p>Work requests that require this one to complete successfully before they can run.</p>
    {% widget work_request_reverse_dependencies %}
{% else %}
    <p>This work request is not required by any other one.</p>
{% endif %}
