{% extends base_template %}
{% load debusine %}
{% block content %}
    <div id="suite-summary">
        <h2>{{ suite.name }}</h2>
        {% if suite.description %}<div id="suite-description" class="mb-4">{{ suite.description }}</div>{% endif %}
    </div>
    {% if suite.architectures %}
        <div id="suite-architectures" class="mb-4">
            <h3>Architectures</h3>
            <p>{{ suite.architectures|join:" " }}</p>
        </div>
    {% endif %}
    {% if suite.release_fields %}
        <div id="suite-release-fields" class="mb-4">
            <h3>Release fields</h3>
            <dl class="row">
                {% for name, value in suite.release_fields.items %}
                    <dt class="col-sm-3 col-lg-2">{{ name }}</dt>
                    <dd class="col-sm-9 col-lg-10">
                        {{ value }}
                    </dd>
                {% endfor %}
            </dl>
        </div>
    {% endif %}
    <div id="suite-apt" class="mb-4">
        <h3>APT source</h3>
        <pre><code>{{ suite.as_deb822_source }}</code></pre>
    </div>
    {% with required_apt_sources=suite.required_suites_as_deb822_sources %}
        {% if required_apt_sources %}
            <div id="suite-required-apt" class="mb-4">
                <h3>Required APT sources</h3>
                <p class="text-muted">The following APT sources are also required by this suite.</p>
                <pre><code>{{ required_apt_sources }}</code></pre>
            </div>
        {% endif %}
    {% endwith %}
{% endblock %}
