{% load debusine %}
{% load artifacts %}
<div class="d-flex justify-content-between">
    <div>
        <span class="bi bi-{{ value.category|artifact_icon_name }}"
              title="{{ value.category }}"></span>
        {% if artifact %}
            <a href="{{ artifact.get_absolute_url }}">{{ artifact.get_label }}</a>
        {% else %}
            <span class="expired">#{{ value.artifact_id }}</span>
        {% endif %}
    </div>
    {% if artifact %}
        {% ui artifact as artifact_ui %}
        {% if artifact_ui.ui_shortcuts %}
            <div class="btn-group btn-group-sm"
                 role="group"
                 aria-label="Artifact shortcuts">
                {% for shortcut in artifact_ui.ui_shortcuts %}
                    {% widget shortcut %}
                {% endfor %}
            </div>
        {% endif %}
    {% endif %}
</div>
